/* See LICENSE file for copyright and license details. */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <net/if.h>
#ifdef __linux__
#include <linux/if_link.h>
#else
#include <sys/sysctl.h>
#include <net/if_dl.h>
#include <net/route.h>
#endif

#include <err.h>
#include <errno.h>
#include <ifaddrs.h>
#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "arg.h"

char *argv0;
int Bflag;
int tflag;

void
scan(char *ifname)
{
	struct ifaddrs *ifas, *ifa;
	int found = 0;

	if (getifaddrs(&ifas) < 0)
		err(1, "getifaddrs");
	if (*ifname) {
		for (ifa = ifas; ifa; ifa = ifa->ifa_next) {
			if (!strncmp(ifname, ifa->ifa_name, IF_NAMESIZE)) {
				found = 1;
				break;
			}
		}
	} else {
		for (ifa = ifas; ifa; ifa = ifa->ifa_next) {
			if (ifa->ifa_flags & IFF_LOOPBACK)
				continue;
			if (ifa->ifa_flags & IFF_RUNNING) {
				if (ifa->ifa_flags & IFF_UP) {
					strncpy(ifname, ifa->ifa_name, IF_NAMESIZE);
					ifname[IF_NAMESIZE - 1] = '\0';
					found = 1;
					break;
				}
			}
		}
	}
	freeifaddrs(ifas);
	if (!found)
		errx(1, "no usable interface found");
}

#ifdef __linux__
void
sample(char *ifname, uint64_t *rxbytes, uint64_t *txbytes,
       uint64_t *rxpps, uint64_t *txpps)
{
	struct ifaddrs *ifas, *ifa;
	struct rtnl_link_stats *stats = NULL;

	if (getifaddrs(&ifas) < 0)
		err(1, "getifaddrs");
	for (ifa = ifas; ifa; ifa = ifa->ifa_next) {
		if (strcmp(ifa->ifa_name, ifname))
			continue;
		if (!ifa->ifa_data || !ifa->ifa_addr)
			break;
		if (ifa->ifa_addr->sa_family == AF_PACKET) {
			stats = ifa->ifa_data;
			*rxbytes = stats->rx_bytes;
			*txbytes = stats->tx_bytes;
			*rxpps = stats->rx_packets;
			*txpps = stats->tx_packets;
			break;
		}
	}
	freeifaddrs(ifas);
	if (!stats)
		errx(1, "interface %s cannot be sampled", ifname);
}
#else
void
sample(char *ifname, uint64_t *rxbytes, uint64_t *txbytes,
       uint64_t *rxpps, uint64_t *txpps)
{
	int mib[6];
	char *buf, *next;
	size_t size;
	struct if_msghdr *ifm;
	struct sockaddr_dl *sdl = NULL;

	*rxbytes = *txbytes = *rxpps = *txpps = 0;
	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;
	mib[3] = 0;
	mib[4] = NET_RT_IFLIST;
	mib[5] = 0;
	if (sysctl(mib, 6, NULL, &size, NULL, 0) < 0)
		errx(1, "sysctl failed");
	buf = malloc(size);
	if (!buf)
		err(1, "malloc");
	if (sysctl(mib, 6, buf, &size, NULL, 0) < 0)
		errx(1, "sysctl failed");
	for (next = buf; next < buf + size; next += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)next;
		if (ifm->ifm_type != RTM_NEWADDR) {
			if (ifm->ifm_flags & IFF_UP) {
				sdl = (struct sockaddr_dl *)(ifm + 1);
				if (sdl->sdl_family != AF_LINK)
					continue;
				if (strncmp(sdl->sdl_data, ifname, sdl->sdl_nlen))
					continue;
				*rxbytes = ifm->ifm_data.ifi_ibytes;
				*txbytes = ifm->ifm_data.ifi_obytes;
				*rxpps = ifm->ifm_data.ifi_ipackets;
				*txpps = ifm->ifm_data.ifi_opackets;
				break;
			}
		}
	}
	free(buf);
	if (!sdl)
		errx(1, "interface %s cannot be sampled", ifname);
}
#endif

void
scale(char **suffix, uint64_t *bits)
{
	static char *suffixes[][3] = { { "bps", "Kbps", "Mbps" },
	                               { "Bps", "KBps", "MBps" } };

	if (Bflag)
		*bits /= 8;
	*suffix = suffixes[Bflag][0];
	if (*bits >= 1000 && *bits < 1000 * 1000) {
		*suffix = suffixes[Bflag][1];
		*bits /= 1000;
	} else if (*bits >= 1000 * 1000) {
		*suffix = suffixes[Bflag][2];
		*bits /= 1000 * 1000;
	}
}

void
print(char *ifname, uint64_t rxbits, uint64_t txbits,
      uint64_t rxpps, uint64_t txpps)
{
	char *rxsuffix, *txsuffix;

	if (!tflag) {
		scale(&rxsuffix, &rxbits);
		scale(&txsuffix, &txbits);
		printf("%s: %3llu Rx %4s %3llu Tx %4s %5llu Rx pps %5llu Tx pps\n",
		       ifname,
		       (unsigned long long)rxbits,
		       rxsuffix,
		       (unsigned long long)txbits,
		       txsuffix,
		       (unsigned long long)rxpps,
		       (unsigned long long)txpps);
	} else {
		printf("%llu %llu %llu %llu\n",
		       (unsigned long long)rxbits,
		       (unsigned long long)txbits,
		       (unsigned long long)rxpps,
		       (unsigned long long)txpps);
	}
	fflush(stdout);
}

int
getmonotime(struct timeval *tv)
{
	struct timespec ts;
	int ret;

	ret = clock_gettime(CLOCK_MONOTONIC, &ts);
	if (ret < 0)
		return -1;
	tv->tv_sec = ts.tv_sec;
	tv->tv_usec = ts.tv_nsec / 1000;
	return ret;
}

void
ms2tv(struct timeval *tv, long ms)
{
	tv->tv_sec = ms / 1000;
	tv->tv_usec = (ms % 1000) * 1000;
}

long
tv2ms(struct timeval *tv)
{
	return (tv->tv_sec * 1000) + ((tv->tv_usec + 500) / 1000);
}

int
tvsleep(struct timeval *tv)
{
	struct timespec ts;

	ts.tv_sec = tv->tv_sec;
	ts.tv_nsec = tv->tv_usec * 1000;
	return nanosleep(&ts, NULL);
}

void
loop(char *ifname, long count, struct timeval *delay)
{
	struct timeval old, now, diff;
	uint64_t oldrxbytes, rxbytes, diffrxbits;
	uint64_t oldtxbytes, txbytes, difftxbits;
	uint64_t oldrxpps, rxpps, diffrxpps;
	uint64_t oldtxpps, txpps, difftxpps;
	long n = 0;

	getmonotime(&old);
	sample(ifname, &oldrxbytes, &oldtxbytes, &oldrxpps, &oldtxpps);
	for (;;) {
		tvsleep(delay);
		sample(ifname, &rxbytes, &txbytes, &rxpps, &txpps);
		getmonotime(&now);
		timersub(&now, &old, &diff);
		getmonotime(&old);
		diffrxbits = (rxbytes - oldrxbytes) * 8;
		difftxbits = (txbytes - oldtxbytes) * 8;
		diffrxbits = diffrxbits * 1000 / tv2ms(&diff);
		difftxbits = difftxbits * 1000 / tv2ms(&diff);
		diffrxpps = (rxpps - oldrxpps) * 1000 / tv2ms(&diff);
		difftxpps = (txpps - oldtxpps) * 1000 / tv2ms(&diff);
		print(ifname, diffrxbits, difftxbits, diffrxpps, difftxpps);
		oldrxbytes = rxbytes, oldtxbytes = txbytes;
		oldrxpps = rxpps, oldtxpps = txpps;
		if (count && ++n >= count)
			break;
	}
}

void
usage(void)
{
	fprintf(stderr, "usage: %s [-B | -t] [-c count] [-d delay] [-i interface]\n", argv0);
	exit(1);
}

int
main(int argc, char *argv[])
{
	char ifname[IF_NAMESIZE] = "";
	char *end;
	struct timeval tv;
	long count = 0, delay = 1000;

	ARGBEGIN {
	case 'B':
		Bflag = 1;
		break;
	case 'c':
		errno = 0;
		count = strtol(EARGF(usage()), &end, 10);
		if (*end != '\0' || errno)
			errx(1, "invalid count");
		break;
	case 'd':
		errno = 0;
		delay = strtol(EARGF(usage()), &end, 10);
		if (*end != '\0' || errno)
			errx(1, "invalid delay");
		break;
	case 'i':
		strncpy(ifname, EARGF(usage()), sizeof(ifname));
		ifname[IF_NAMESIZE - 1] = '\0';
		break;
	case 't':
		tflag = 1;
		break;
	default:
		usage();
	} ARGEND

	ms2tv(&tv, delay);
	scan(ifname);
	loop(ifname, count, &tv);
	return 0;
}
